/*
**	Sambar Server JAVA Engine Implementation
**
**	Confidential Property of Tod Sambar
**	(c) Copyright Tod Sambar 2002
**	All rights reserved.
*/
package com.sambar.javaeng;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.AbstractList;
import java.util.ArrayList;
import java.util.NoSuchElementException;
import java.util.Stack;
import java.util.Vector;
import java.util.Iterator;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.servlet.UnavailableException;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;


public class SambarXmlParser 
{
    private SAXParser 	parser;
    
    public SambarXmlParser()
		throws UnavailableException
    {
        try
        {
        	SAXParserFactory factory = SAXParserFactory.newInstance();
            factory.setValidating(false);
            factory.setNamespaceAware(false);
            parser = factory.newSAXParser();
        }
        catch (Exception e)
        {
            throw new UnavailableException("SambarXmlParser faillure: " +
				e.toString());
        }
    }
    
    public synchronized Node parse(InputSource source)
        throws IOException, SAXException
    {
        Handler handler = new Handler();
        XMLReader reader = parser.getXMLReader();
        reader.setContentHandler(handler);
		reader.setErrorHandler(handler);
		reader.setEntityResolver(handler);
        parser.parse(source, handler);

        if (handler.error != null)
            throw handler.error;

        Node doc = (Node)handler.top.get(0);
        handler.clear();
        return doc;
    }
    
    public synchronized Node parse(String url)
        throws IOException, SAXException
    {
        return parse(new InputSource(url));
    }
    
    public synchronized Node parse(File file)
        throws IOException, SAXException
    {
        return parse(new InputSource(file.toURL().toString()));
    }
    
    /* ------------------------------------------------------------ */
    private class Handler extends DefaultHandler
    {
        Node top = new Node(null,null,null);
        SAXParseException error;
        private Node context = top;

        void clear()
        {
            this.top=null;
            this.error=null;
            this.context=null;
        }
        
        public void startElement (String uri, String localName, String qName, 
			Attributes attrs)
            throws SAXException
        {
            String name = (uri==null || uri.equals("")) ? qName : localName;
            Node node = new Node(context,name,attrs);
            context.add(node);
            context = node;
        }

        public void endElement (String uri, String localName, String qName)
            throws SAXException
        {
            context = context.parent;
        }

        public void ignorableWhitespace (char buf [], int offset, int len)
            throws SAXException
        {
        }

        public void characters (char buf [], int offset, int len)
            throws SAXException
        {
			String s = new String(buf,offset,len);
            context.add(new String(buf,offset,len));
        }
        
        public void warning(SAXParseException ex)
        {
            // Ignore warnings at this time.
			System.out.println("SAX Parser warning " + ex.toString());
        }
                
        public void error(SAXParseException ex)
            throws SAXException
        {
			System.out.println("SAX Parser error " + ex.toString());

            // Save error 
            if (error == null)
                error = ex;
        }
        
        public void fatalError(SAXParseException ex)
            throws SAXException
        {
			System.out.println("SAX Parser fatal error " + ex.toString());
            error = ex;
            throw ex;
        }	    

        private String getLocationString(SAXParseException ex)
        {
            return ex.getSystemId()+
                " line:" + ex.getLineNumber()+
                " col:" + ex.getColumnNumber();
        }
        
        public InputSource resolveEntity(String pid, String sid)
        {
            return null;
        }
    }
    
    /* ------------------------------------------------------------ */
    public static class Attribute
    {
        private String name;
        private String value;
        Attribute(String n,String v)
        {
            name=n;
            value=v;
        }
        public String getName() 
		{
			return name;
		}
        public String getValue() 
		{
			return value;
		}
    }
    
    /* ------------------------------------------------------------ */
    public static class Node extends AbstractList
    {
        Node parent;
        private ArrayList list;
        private String tag;
        private SambarXmlParser.Attribute[] attrs;
        private boolean lastString=false;
        
        Node(Node parent,String tag, Attributes saxattrs)
        {
            this.parent = parent;
            this.tag = tag;

            if (saxattrs != null)
            {
                attrs = new SambarXmlParser.Attribute[saxattrs.getLength()];
                for (int i = 0; i < saxattrs.getLength(); i++)
				{
		    		String name = saxattrs.getLocalName(i);
		    		if (name == null || name.equals(""))
						name = saxattrs.getQName(i);
                    attrs[i] = new SambarXmlParser.Attribute(name, 
						saxattrs.getValue(i));
                }                    
            }
        }
        
        public Node getParent()
        {
            return parent;
        }  

        public String getTag()
        {
            return tag;
        } 

		public SambarXmlParser.Attribute[] getAttributes()
        {
            return attrs;
        }
        
		public String getAttribute(String name)
        {
            return getAttribute(name, null);
        }
        
		public String getAttribute(String name, String deflt)
        {
            if (attrs==null || name==null)
                return deflt;
            for (int i = 0; i < attrs.length; i++)
			{
                if (name.equals(attrs[i].getName()))
                    return attrs[i].getValue();
			}

            return deflt;
        }
        
		public int size()
        {
            if (list != null)
                return list.size();
            return 0;
        }
        
		public Object get(int i)
        {
            if (list != null)
                return list.get(i);
            return null;
        }
        
		public Node get(String tag)
        {
            if (list != null)
            {
                for (int i = 0; i < list.size(); i++)
                {
                    Object o= list.get(i);
                    if (o instanceof Node)
                    {
                        Node n = (Node)o;
                        if (tag.equals(n.tag))
                            return n;
                    }
                }
            }

            return null;
        }

        public void add(int i, Object o)
        {
            if (list == null)
                list = new ArrayList();

            if (o instanceof String)
            {
                if (lastString)
                {
                    int last = list.size()-1;
                    list.set(last, (String)list.get(last) + o);
                }
                else
				{
                    list.add(i, o);
				}

                lastString = true;
            }
            else
            {
                lastString = false;
                list.add(i, o);
            }
        }	

        public void clear()
        {
            if (list != null)
                list.clear();
            list = null;
        }
        
        public String getString(String tag, boolean tags, boolean trim)
        {
            Node node = (Node)get(tag);
            if (node == null)
                return null;

            String s = node.toString(tags);
            if (s != null && trim)
                s = s.trim();

            return s;
        }

        public synchronized String toString()
        {
            return toString(true);
        }
        
		public synchronized String toString(boolean tag)
        {
            StringBuffer buf = new StringBuffer();
            synchronized(buf)
            {
                toString(buf, tag);
                return buf.toString();
            }
        }
        
		public synchronized String toString(boolean tag, boolean trim)
        {
            String s = toString(tag);
            if (s != null && trim)
                s = s.trim();

            return s;
        }
        
        private synchronized void toString(StringBuffer buf,boolean tag)
        {
            if (tag)
            {
                buf.append("<");
                buf.append(tag);
            
				if (attrs!=null)
				{
					for (int i = 0; i < attrs.length; i++)
					{
						buf.append(' ');
						buf.append(attrs[i].getName());
						buf.append("=\"");
						buf.append(attrs[i].getValue());
						buf.append("\"");
		    		}
				}
            }

            if (list != null)
            {
                if (tag)
                    buf.append(">");
                
                for (int i = 0; i < list.size();i++)
                {
                    Object o= list.get(i);
                    if (o == null)
                        continue;
                    if (o instanceof Node)
                        ((Node)o).toString(buf,tag);
                    else
                        buf.append(o.toString());
                }

                if (tag)
                {
                    buf.append("</");
                    buf.append(tag);
                    buf.append(">");
                }
            }
            else if (tag)
			{
                buf.append("/>");
			}
        }

        public Iterator iterator(final String tag)
        {
            return new Iterator()
            {
                int c = 0;
                Node node = null;

                public boolean hasNext()
                {
                    if (node != null)
                        return true;

                    while ((list != null) && (c < list.size()))
                    {
                        Object o = list.get(c);
                        if (o instanceof Node)
                        {
                            Node n = (Node)o;

                            if (tag.equals(n.tag))
                            {
                                node = n;
                                return true;
                            }
                        }
                        c++;
                    }
                    return false;
                }

                public Object next()
                {
                    try
                    {
                        if (hasNext())
                            return node;

                        throw new NoSuchElementException();
                    }
                    finally
                    {
                        node=null;
                        c++;
                    }
                }

                public void remove()
                {
                    throw new UnsupportedOperationException("Not supported");
                }

            };
        }
    }
}
